﻿// $Id: $
package com.nesterovskyBros.test;

import java.lang.ref.WeakReference;

import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.TimeUnit;

/**
 * An object cache storing objects with expiration.
 * Note that values are cached until their expiration points. An expiration 
 * point is updated after invoke of get() method.
 * @param <K> a key type.
 * @param <V> a value type.
 */
public class Cache<K, V>
{
  /**
   * Default expiration interval (10 minutes).
   */
  public static final int defaultExpirationInterval = 10 * 60 * 1000;
  
  /**
   * Creates a {@link Cache} instance.
   */
  public Cache()
  {
    this.defaultExpiration = defaultExpirationInterval;
  }
  
  /**
   * Creates a {@link Cache} instance.
   * @param defaultExpiration a default expiration interval.
   */
  public Cache(long defaultExpiration)
  {
    this.defaultExpiration = defaultExpiration;
  }
  
  /**
   * Gets a cached object for a specified key.
   * @param key a key of the object.
   * @return an object for a key, or null if object is not in the cache.
   */
  public V get(K key)
  {
    Item item = cache.get(key);
    
    if (item == null)
    {
      return null;
    }
    else
    {
      item.lastAccess = System.currentTimeMillis();
      
      return item.value;
    }
  }
  
  /**
   * Puts an object into the cache with default value of the sliding expiration.
   * @param key a key of the object (cannot be null).
   * @param value value a value to put. If value is null then object 
   *   corresponding to the specified key is removed from the cache.
   */
  void put(K key, V value)
  {
    put(
      key,
      value,
      defaultExpiration,
      true);
  }
  
  /**
   * Puts an object into the cache.
   * @param key a key of the object (cannot be null).
   * @param value a value to put. If value is null then object corresponding to 
   *   the specified key is removed from the cache.
   * @param expiration expiration an expiration in milliseconds.
   * @param slidingExpiration true for sliding, and false for absolute 
   *   expiration.
   */
  void put(
    K key,
    V value,
    long expiration,
    boolean slidingExpiration)
  {
    long now = System.currentTimeMillis();
    
    if (value == null)
    {
      Item previous = cache.remove(key);
      
      if (previous != null)
      {
        previous.lastAccess = now - previous.expiration;
      }
    }
    else
    {
      Item item = new Item();
      
      item.key = key;
      item.value = value;
      item.expiration = expiration;
      item.slidingExpiration = slidingExpiration;
      item.lastAccess = now;
      
      Item previous = cache.put(key, item);
      
      if (previous != null)
      {
        previous.lastAccess = now - previous.expiration;
      }
      
      Task<K, V> task = new Task<K, V>();
      
      task.item = new WeakReference<Item>(item);
      executor.schedule(
        task,
        item.expiration,
        TimeUnit.MILLISECONDS);
    }
  }
  
  /**
   * Scheduler task.
   * Note that a {@link Task} does not prevent a {@link Cache} to be collected.
   */
  private static class Task<K, V> implements Runnable
  {
    /**
     * Item reference.
     */
    public WeakReference<Cache<K, V>.Item> item;
    
    /**
     * Invalidates an item.
     */
    public void run()
    {
      Cache<K, V>.Item item = this.item.get();
      
      if (item != null)
      {
        long now = System.currentTimeMillis();
        long delta = item.lastAccess + item.expiration - now;
        
        if (item.slidingExpiration && delta > 0)
        {
          executor.schedule(
            this,
            delta,
            TimeUnit.MILLISECONDS);
        }
        else
        {
          item.cache().remove(item.key, item);
        }
      }
    }
  }
  
  /**
   * An item of the storage map.
   */
  private class Item
  {
    /**
     * Gets the storage map.
     * @return the storage map.
     */
    public ConcurrentHashMap<K, Item> cache()
    {
      return cache;
    }
    
    /**
     * Object key.
     */
    public K key;
    
    /**
     * Object itself.
     */
    public V value;
    
    /**
     * Expiration interval.
     */
    public long expiration;
    
    /**
     * true for sliding, and false for absolute expiration.
     */
    public boolean slidingExpiration;
    
    /**
     * Last access instant.
     */
    public volatile long lastAccess;
  }
  
  /**
   * A default expiration value.
   */
  private final long defaultExpiration;
  
  /**
   * Objects storage.
   */
  private final ConcurrentHashMap<K, Item> cache = new ConcurrentHashMap<K,
     Item>();
  
  /**
   * An executor reference to clean caches.
   */
  private static final ScheduledThreadPoolExecutor executor;
  
  /**
   * Sets up a global executor to clean caches.
   */
  static
  {
    executor = new ScheduledThreadPoolExecutor(
      1,
      new ThreadFactory()
      {
        public Thread newThread(Runnable target)
        {
          Thread thread = new Thread(target, "Cache cleaner");
          
          thread.setDaemon(true);
          thread.setPriority(Thread.MAX_PRIORITY);
          
          return thread;
        }
      });
    executor.setMaximumPoolSize(1);
    executor.setKeepAliveTime(1, TimeUnit.MINUTES);
    executor.allowCoreThreadTimeOut(true);
  }
}
